# environment setup

const scriptpath = @__DIR__
const datapath = joinpath(@__DIR__, "..", "data")


#using Pkg
#Pkg.activate(scriptpath)
#Pkg.instantiate()

#    function savelog(df, name)
#        CSV.write(joinpath(datapath, "$(name)_energies.csv"), df)
#        #display(df)
#        df2 = DataFrame([a => a == :k ? b[1:end-1] : abs.(diff(b)) for (a, b) in pairs(eachcol(df))])
#        #display(df2)
#        CSV.write(joinpath(datapath, "$(name)_energydiffs.csv"), df2)
#    end
# experiments

using Random: MersenneTwister
using LinearAlgebra: Diagonal, I, dot, norm, mul!, diagm
using Statistics: median

using CSV
using Colors: HSV
using DataFrames
using Distributed
@everywhere using DualTVDD:
    DualTVL1ROFOpProblem,
    DualTVDDAlgorithm, ChambolleAlgorithm, NStepAlgorithm, ProjGradAlgorithm,# TVNewtonAlgorithm,
    SurrogateAlgorithm,
    DualTVDDSurrogateAlgorithm,
    init, step!, fetch, fetch_u, gradient, divergence, normB, energy, residual
using DualTVDD: fetch_u
using FileIO
using ImageIO
using OpticalFlowUtils
using FileIO: @format_str, add_format
add_format(format"FLO", b"PIEH", ".flo", [:OpticalFlowUtils])
using Optim
using Plots
using FFTW
using StaticKernels: Kernel, ExtensionReplicate, ExtensionNothing, extend
using LinearMaps
using IterativeSolvers

include("util.jl")
isdefined(Main, :Revise) && Revise.track(joinpath(@__DIR__, "util.jl"))

using .Util

# Problems

function DenoiseProblem(img::AbstractArray{T,d}; λ, β = 0.) where {T,d}
    B = (1 + β) * I
    g = img

    return DualTVL1ROFOpProblem(g, B, λ)
end

function InpaintProblem(img::AbstractArray{T,d}, imgmask::AbstractArray{Bool,d}; λ, β) where {T,d}
    pwop = imgmask .+ β .* Ref(I)
    B = Diagonal(vec(inv.(pwop)))

    ## we grey-out the inpainting area.
    ## since β > 0 this actually matters and should be a sane default
    #g = imgmask .* img .+ .!imgmask .* 0.5
    g = imgmask .* img
    return DualTVL1ROFOpProblem(g, B, λ)
end

function OptFlowProblem(f0::AbstractArray{T,d}, f1::AbstractArray{T,d}; λ, β) where {T,d}
    size(f0) == size(f1) || throw(ArgumentError("mismatched sizes"))

    ∇ = Kernel{ntuple(_->0:1, d)}(gradient)
    ∇f0 = map(∇, extend(f0, ExtensionReplicate()))

    Bval = inv.(∇f0 .* adjoint.(∇f0) .+ β .* Ref(I))
    B = Diagonal(vec(Bval))

    g = .-(f1 .- f0) .* ∇f0

    return DualTVL1ROFOpProblem(g, B, λ)
end

function DFTProblem(img::AbstractArray{T,d}; λ, β=0.) where {T,d}
    T <: Complex || throw(ArgumentError("need `<:Complex` eltype for frequency data"))
    n = length(img)
    B = 1 / (n + β) * I
    g = real.(n * ifft(img))

    return DualTVL1ROFOpProblem(g, B, λ)
end

function GlobalProblem(img::AbstractArray{T,d}, A; λ, β) where {T,d}
    n = length(img)

    B = LinearMap{Float64}(
        b -> cg(A'*A + β * I, b; maxiter=10),
        n, n; issymmetric=true, isposdef=β>0)

    #B = inv(A'*A + β*I)
    #B = (B' + B) / 2
    g = reshape(A' * A * vec(img), size(img))

    return DualTVL1ROFOpProblem(g, B, λ)
end

# Util

toimg(arr) = Gray.(clamp.(arr, 0., 1.))
loadimg(x) = Float64.(load(x))
saveimg(io, x) = save(io, toimg(x))


function run_algs(dataf, algs; maxiters)
    flatten(x; kwargs...) =
        (;kwargs..., (Symbol(a, "_", c) => d for (a, b) in pairs(x) for (c, d) in pairs(b))...)

    df = DataFrame()
    states = map(init, algs)
    push!(df, flatten(map(dataf, states); k=0))

    # log-skipping parameter
    a = 20
    for k in 1:maxiters
        round(a*log(k+1)) - round(a*log(k)) == 0 && continue
        push!(df.k, k)
    end
    for (algname, st) in pairs(states)
        for k in 1:maxiters
            step!(st)
            round(a*log(k+1)) - round(a*log(k)) == 0 && continue
            stepdata = dataf(st)
            for (dataname, data) in pairs(stepdata)
                push!(getproperty(df, Symbol(algname, "_", dataname)), data)
            end
        end
    end

    return df, states
end

comparison_algs(prob; τ, M, overlap, ninner) = (
    glob = NStepAlgorithm(ChambolleAlgorithm(prob; τ), ninner),
    ddseq = DualTVDDAlgorithm(prob; M, overlap, ninner, parallel=false, σ=1., subalg=x->ChambolleAlgorithm(x; τ)),
    ddseqav = DualTVDDAlgorithm(prob; M, overlap, ninner, parallel=false, σ=0.25, subalg=x->ChambolleAlgorithm(x; τ)),
    ddpar = DualTVDDAlgorithm(prob; M, overlap, ninner, parallel=true, σ=0.25, subalg=x->ChambolleAlgorithm(x; τ)))

function calc_energymin(prob, alg; maxiters)
    st = init(alg)
    for _ in 1:maxiters
        step!(st)
    end
    return energy(fetch(st), prob)
end

# Experiment Suite

function compare_algs(prob; τ, M, overlap, maxiters, ninner)
    algs = comparison_algs(prob; τ, M, overlap, ninner)

    df, states = run_algs(algs; maxiters=maxiters ÷ 10) do state
        return (
            energy = energy(fetch(state), prob),
            residual = residual(fetch(state), prob))
    end

    #filter!(:k => k -> k < maxiters / 10, df)

    energymin = calc_energymin(prob, algs.glob; maxiters)
    for x in names(df)
        contains(x, r"_energy$") || continue
        df[!,Symbol(x)] .-= energymin
    end

    return df, states, energymin
end

function run_while(fwhile, alg)
    st = init(alg)
    k = 0
    while fwhile(st, k)
        k += 1
        step!(st)
    end
    return st
end

function suite(::Path{(:convergence, :test)}, input, output)
    #fo = loadimg(input("input_original.png"))

    #λ = 0.1
    #β = 0.
    #maxiters = 10
    #ninner = 10

    #M = (2,2)
    #overlap = (5,5)

    #noise_σ = 0.1

    #f = clamp.(fo .+ noise_σ * randn(Float64, size(fo)), 0., 1.)
    #saveimg(output("input_noisy.png"), f)

    #prob = DenoiseProblem(f; λ, β)

    fo = f0 = loadimg(input("input_f0.png"))
    f1 = loadimg(input("input_f1.png"))

    λ = 0.002
    β = 0.01

    maxiters = 100
    ninner = 1000

    M = (4,4)
    overlap = (5,5)

    prob = OptFlowProblem(f0, f1; λ, β)

    algs = (;
        #glob = ChambolleAlgorithm(prob),
        #surrogate = SurrogateAlgorithm(prob; ninner=100, τ=2*normB(prob)+eps(normB(prob)),
        #                               subalg=x->ChambolleAlgorithm(x)),

        surrogate_ddseq = SurrogateAlgorithm(prob; ninner=1, τ = normB(prob) + eps(normB(prob)),
            subalg = x -> DualTVDDAlgorithm(x; M, overlap, ninner, parallel = false, σ = 1.,
                subalg = x -> ChambolleAlgorithm(x))),
        ddseq_surrogate = DualTVDDAlgorithm(prob; M, overlap, ninner, parallel = false, σ = 1.,
            subalg = x -> SurrogateAlgorithm(x; ninner=1, τ = normB(x) + eps(normB(x)),
                subalg = x -> ChambolleAlgorithm(x))),
        #ddseq = DualTVDDAlgorithm(prob; M, overlap, ninner, parallel = false, σ = 1.,
        #    subalg = x -> ChambolleAlgorithm(x)),
       )

    #algs = comparison_algs(prob; τ, M, overlap, ninner)

    df, states = run_algs(algs; maxiters) do state
        return (
            energy = energy(fetch(state), prob),
            residual = residual(fetch(state), prob))
    end
    display(df)

    CSV.write(output("energies.csv"), df)
    #saveimg(output("output_glob.png"), fetch_u(states.glob))
    savedata(output("data.tex"); lambda=λ, beta=β, maxiters,
        width=size(fo, 2), height=size(fo, 1))
end

function suite(::Path{(:scaling, :opticalflow)}, input, output)
    fo = f0 = loadimg(input("frame10.png"))
    f1 = loadimg(input("frame11.png"))
    d = 2

    λ = 0.01
    β = 0.001

    ninner = 300
    Mdir = 2 * floor(Int, sqrt(nworkers()))
    M = (Mdir, Mdir)
    overlap = (5, 5)

    stopenergy = 130.
    ntimings = 3

    prob = OptFlowProblem(f0, f1; λ, β)
    #return prob

    galg = ChambolleAlgorithm(prob)
    dalg(workers) =
        DualTVDDAlgorithm(prob;
            workers, M, overlap, ninner, parallel=false, σ=1.,
            subalg = x -> ChambolleAlgorithm(x))

    function timeit(alg)
        # precompile
        run_while((_, _) -> false, alg)

        times = map(1:ntimings) do _
            return @elapsed run_while(alg) do st, k
                # min ~107
                return energy(fetch(st), prob) > stopenergy
            end
        end
        return minimum(times)
    end

    tg = timeit(galg)

    ws = workers()
    df = DataFrame()
    for np in 0:Mdir÷2
        nw = 2^np
        push!(df, (
            nworkers = nw,
            time = timeit(dalg(ws[1:nw]))))
    end

    display(df)

    CSV.write(output("timings.csv"), df)
    #saveimg(output("output_glob.png"), fetch_u(states.glob))
    savedata(output("data.tex"); lambda=λ, beta=β, Mdir,
        ntimings, stopenergy, ninner,
        width=size(fo, 2), height=size(fo, 1))
end

function prob_outerinner(input)
end

suite(::Path{(:surrogate, :outerinner)}, input, output) =
    _surrogate_outerinner(input, output, false)
suite(::Path{(:surrogate, :outerinner_ref)}, input, output) =
    _surrogate_outerinner(input, output, true)

function _surrogate_outerinner(input, output, ref)
    fo = f0 = loadimg(input("frame10.png"))
    f1 = loadimg(input("frame11.png"))

    λ = 0.01
    β = 0.001
    prob = OptFlowProblem(f0, f1; λ, β)

    maxiters = 100
    ninner = 50
    minenergy_maxiters = 10 * maxiters * ninner

    M = (4,4)
    overlap = (50,50)

    # set B = I for reference
    if ref
        prob.B.diag .= Ref(Matrix(I, 2, 2))
    end

    energymin = calc_energymin(prob, ChambolleAlgorithm(prob); maxiters=minenergy_maxiters)

    algs = (;
        surrogate_ddseq = SurrogateAlgorithm(prob; ninner=1, τ = normB(prob) + eps(normB(prob)),
            subalg = x -> DualTVDDAlgorithm(x; M, overlap, ninner, parallel = false, σ = 1.,
                subalg = x -> ChambolleAlgorithm(x))),
        #surrogate_ddpar = SurrogateAlgorithm(prob; ninner=1, τ = normB(prob) + eps(normB(prob)),
        #    subalg = x -> DualTVDDAlgorithm(x; M, overlap, ninner, parallel = true, σ = 0.25,
        #        subalg = x -> ChambolleAlgorithm(x))),
        ddseq_surrogate = DualTVDDAlgorithm(prob; M, overlap, ninner=1, parallel = false, σ = 1.,
            subalg = x -> SurrogateAlgorithm(x; ninner, τ = normB(x) + eps(normB(x)),
                subalg = x -> ChambolleAlgorithm(x))),
        #ddpar_surrogate = DualTVDDAlgorithm(prob; M, overlap, ninner=1, parallel = true, σ = 0.25,
        #    subalg = x -> SurrogateAlgorithm(x; ninner, τ = normB(x) + eps(normB(x)),
        #        subalg = x -> ChambolleAlgorithm(x))),
       )

    df, states = run_algs(algs; maxiters) do state
        return (
            energy = energy(fetch(state), prob) - energymin,
            #residual = residual(fetch(state), prob),
           )
    end
    display(df)

    CSV.write(output("energies.csv"), df)
    #saveimg(output("output_glob.png"), fetch_u(states.glob))
    savedata(output("data.tex"); lambda=λ, beta=β, minenergy_maxiters, maxiters, ninner, M, overlap,
        width=size(fo, 2), height=size(fo, 1))
end

function suite(::Path{(:convergence, :denoising)}, input, output)
    fo = loadimg(input("input_original.png"))

    λ = 0.1
    β = 0.
    maxiters = 10000
    ninner = 100

    M = (2,2)
    overlap = (5,5)

    noise_σ = 0.1

    f = clamp.(fo .+ noise_σ * randn(Float64, size(fo)), 0., 1.)
    saveimg(output("input_noisy.png"), f)

    prob = DenoiseProblem(f; λ, β)

    τ = inv(8 * normB(prob))
    df, states, energymin = compare_algs(prob; τ, M, overlap, maxiters, ninner)

    CSV.write(output("energies.csv"), df)
    saveimg(output("output_glob.png"), fetch_u(states.glob))
    savedata(output("data.tex"); lambda=λ, beta=β, tau=τ, maxiters, noise_sigma=noise_σ, energymin,
        width=size(fo, 2), height=size(fo, 1))
end

function suite(::Path{(:convergence, :inpainting)}, input, output)
    fo = loadimg(input("input_original.png"))

    λ = 0.05
    β = 0.001
    maxiters = 100000
    ninner = 100

    M = (2,2)
    overlap = (5,5)

    f_mask = rand(Float64, size(fo)) .> 0.5

    prob = InpaintProblem(fo, f_mask; λ, β)
    saveimg(output("input_mask.png"), f_mask)
    saveimg(output("input_g.png"), prob.g)

    τ = inv(8 * normB(prob))
    df, states, energymin = compare_algs(prob; τ, M, overlap, maxiters, ninner)

    CSV.write(output("energies.csv"), df)
    saveimg(output("output_glob.png"), fetch_u(states.glob))
    savedata(output("data.tex"); lambda=λ, beta=β, tau=τ, maxiters, energymin,
        width=size(fo, 2), height=size(fo, 1))
end

function suite(::Path{(:convergence, :opticalflow)}, input, output)
    f0 = loadimg(input("input_f0.png"))
    f1 = loadimg(input("input_f1.png"))

    λ = 0.002
    β = 0.001

    maxiters = 100000
    ninner = 100

    M = (2,2)
    overlap = (5,5)

    prob = OptFlowProblem(f0, f1; λ, β)

    τ = inv(8 * normB(prob))
    df, states, energymin = compare_algs(prob; τ, M, overlap, maxiters, ninner)

    CSV.write(output("energies.csv"), df)

    cflow = colorflow(fetch_u(first(states)); maxflow=4.5/(584/48))
    #p = plot(cflow); display(p)
    save(output("flow.png"), cflow)

    savedata(output("data.tex"); lambda=λ, beta=β, tau=τ, maxiters, energymin,
        width=size(f0, 2), height=size(f0, 1))
end

function dwt!(y, x)
    a = sqrt(1 / 2)
    n = length(x)
    n <= 1 && return (y .= x)
    # use second half of y as scratch space
    for i in 1:n÷2
        j = 2*(i-1)+1
        y[n÷2+i] = a * x[j] + a * x[j+1]
    end
    dwt!(view(y, 1:n÷2), view(y, n÷2+1:2*(n÷2)))
    for i in 1:n÷2
        j = 2*(i-1)+1
        y[n÷2+i] = a * x[j] - a * x[j+1]
    end
    isodd(n) && (y[n] = x[n])
    return y
end

function idwt!(y, x)
    a = sqrt(1 / 2)
    n = length(x)
    n <= 1 && return (y .= x)
    isodd(n) && (y[n] = x[n])
    # use second half of y as scratch space
    idwt!(view(y, n÷2+1:2*(n÷2)), view(x, 1:n÷2))
    for i in 1:n÷2
        j = 2*(i-1)+1
        y[j]   = a * y[n÷2+i] + a * x[n÷2+i]
        y[j+1] = a * y[n÷2+i] - a * x[n÷2+i]
    end
    return y
end

# TODO: non-allocating
function mdwt!(y, x)
    d = ndims(x)
    s = size(x)
    any(s .<= 1) && return (y .= x)

    # transformation matrix
    M = [sqrt(1/2^d)*(-1)^count_ones(i&j) for i in 0:2^d-1, j in 0:2^d-1]

    offset = s.÷2
    firsthalf = Base.OneTo.(offset)
    for i in Iterators.product(firsthalf...)
        j = @. 2*(i-1)+1

        src_ix = ntuple(k->j[k]:1+j[k], d)
        dst_ix = ntuple(k->i[k]:offset[k]:offset[k]+i[k], d)

        src = vec(view(x, src_ix...))
        dst = vec(view(y, dst_ix...))
        mul!(dst, M, src)
    end

    mdwt!(view(y, firsthalf...), y[firsthalf...])

    for k in 1:d
        iseven(s[k]) && continue
        selectdim(y, k, s[k]) .= selectdim(x, k, s[k])
    end

    return y
end

# TODO: non-allocating
function imdwt!(y, x)
    d = ndims(x)
    s = size(x)
    any(s .<= 1) && return (y .= x)

    offset = s .÷ 2
    firsthalf = Base.OneTo.(offset)

    # transformation matrix
    M = [sqrt(1 / 2^d)*(-1)^count_ones(i&j) for i in 0:2^d-1, j in 0:2^d-1]

    for k in 1:d
        iseven(s[k]) && continue
        selectdim(y, k, s[k]) .= selectdim(x, k, s[k])
    end

    xt = copy(x)
    imdwt!(view(xt, firsthalf...), x[firsthalf...])

    for i in Iterators.product(firsthalf...)
        j = @. 2*(i-1)+1

        src_ix = ntuple(k->i[k]:offset[k]:offset[k]+i[k], d)
        dst_ix = ntuple(k->j[k]:1+j[k], d)

        src = vec(view(xt, src_ix...))
        dst = vec(view(y, dst_ix...))
        mul!(dst, M, src)
    end

    return y
end
mdwt(x) = mdwt!(similar(x), x)
imdwt(x) = imdwt!(similar(x), x)

function suite(::Path{(:global, :basic)}, input, output)
    f = loadimg(input("input_original.png"))
    #f = loadimg(input("frame10_small.png"))
    ax = axes(f)

    seed = 42
    rng = MersenneTwister(seed)

    λ = 0.02
    β = 1e-3

    #maxiters = 100000
    maxiters = 100
    ninner = 1000

    M = (4,4)
    overlap = (5,5)

    n = length(f)
    wr = rand(rng, 1:n, n÷2)
    #wr = n÷4:n÷2

    A = LinearMap{Float64}(
        (y, x) -> (mdwt!(reshape(y, ax), reshape(x, ax)); y[wr] .= 0; y),
        (y, x) -> (x[wr] .= 0; imdwt!(reshape(y, ax), reshape(x, ax))),
        # no longer symmetric
        n, n, issymmetric=false, isposdef=false)

    #A = rand(rng, n, n) ./ n
    #A = diagm(rand(rng, n))

    #A = LinearMap{Float64}(x -> x, x -> x, n, n, issymmetric=true, isposdef=true)

    prob = GlobalProblem(f, A; λ, β)


    #τ = inv(8 * normB(prob))

    algref = ChambolleAlgorithm(prob, τ=β/8)
    st = init(algref)
    u0 = fetch_u(st)

    #alg = ChambolleAlgorithm(prob, τ=β/8)

    τ = 2/β
    alg = DualTVDDSurrogateAlgorithm(prob; M, overlap, ninner, τ, parallel=false)

    #return alg
    #alg = DualTVDDAlgorithm(prob; M, overlap, ninner)

    #display(fetch(st))
    #display(fetch_u(st))
    #step!(st)
    #display(fetch(st))
    #display(fetch_u(st))
    #step!(st)
    #display(fetch(st))
    #display(fetch_u(st))
    #return

    df, states = run_algs([alg]; maxiters=maxiters) do state
        return (
            energy = energy(fetch(state), prob),
            residual = residual(fetch(state), prob))
    end

    #energymin = calc_energymin(prob, algref; maxiters)
    #for x in names(df)
    #    contains(x, r"_energy$") || continue
    #    df[!,Symbol(x)] .-= energymin
    #end

    CSV.write(output("energies.csv"), df)

    u = fetch_u(states[1])
    #display(u)
    p = plot(plot(toimg(f)), plot(toimg(u0)), plot(toimg(u)))

    saveimg(output("input_orig.png"), f)
    saveimg(output("input_g.png"), prob.g)
    saveimg(output("output_u.png"), u)


    display(p)


    savedata(output("data.tex"); lambda=λ, beta=β, tau=τ, maxiters, ninner,
        width=size(f, 2), height=size(f, 1))
    return prob
end


function halve(img)
    all(iseven.(size(img))) || throw(ArgumentError("uneven size"))

    res = similar(img, size(img) .÷ 2)
    fill!(res, zero(eltype(res)))

    for i in CartesianIndices(img)
        j = CartesianIndex((Tuple(i) .+ 1) .÷ 2)
        res[j] += img[i]
    end
    res ./= 2 ^ ndims(img)

    return res
end


"""
    logfilter(dt; a=20)

Reduce dataset such that the resulting dataset would have approximately
equidistant datapoints on a log-scale. `a` controls density.
"""
logfilter(dt; a=20) = filter(:k => k -> round(a*log(k+1)) - round(a*log(k)) > 0, dt)

#function calcmin(f, alg)
#    ctx = init(alg)
#    fmin = f(ctx)
#    while true
#        step!(ctx)
#        fcur = f(ctx)
#        fcur >= fmin && break
#        fmin = min(fmin, fcur)
#    end
#    return fmin
#end
#
#function seq(f; start, length)
#    res = Vector{Float64}(undef, length)
#    res[1] = f(1, start)
#    for i in 2:length
#        res[i] = f(i, res[i-1])
#    end
#    return res
#end
#
#estmin(fk; p = 1) = estmin(axes(fk, 1), fk; p)
#function estmin(k, fk; p = 1, last=length(k)-ceil(Int, 10^(log10(length(k)) * (1-1/3))))
#    k = k[end-last+1:end]
#    fk = fk[end-last+1:end]
#
#    #res = optimize(x -> minimum(fk) .- x[1] <= 0 || x[2] <= 0 ? NaN : sum((log.(fk .- x[1]) .+ 1*log.(k) .+ log(x[2])).^2), [-Inf, nextfloat(0.)], [nextfloat(minimum(fk)), Inf], [minimum(fk)-1, 1.], Fminbox())
#
#    #return res.minimizer[1]
#
#    A = hcat(ones(length(k)), (1 ./ k) .^ p)
#    c0, c1 = A \ fk
#    return c0
#end
