using Distributed: workers
using Outsource: Connector, outsource

#import Serialization
## We override SubArray serialization in order to preserve array data
## references. TODO: this should actually be julia-default!
#function Serialization.serialize(
#    s::Serialization.AbstractSerializer,
#    a::SubArray{T,N,A}) where {T,N,A<:Array}
#
#        Serialization.serialize_any(s, a)
#end


struct DualTVDDAlgorithm{P,d} <: Algorithm{P}
    problem::P

    "number of subdomains in each dimension"
    M::NTuple{d,Int}
    "overlap in pixels per dimension"
    overlap::NTuple{d,Int}
    "use the non-sequential inertia update"
    parallel::Bool
    "inertia parameter (only when parallel)"
    σ::Float64
    "number of inner iterations"
    ninner::Int
    "prob -> Algorithm(::Problem, ...)"
    subalg::Function
    "worker ids used for distributed execution"
    workers::Vector{Int}
    function DualTVDDAlgorithm(problem; M, overlap, parallel=true, σ=parallel ? 1/4 : 1., ninner=10, subalg=x->ChambolleAlgorithm(x), workers=workers())
        if parallel == true && σ > 1/4
            @warn "parallel domain decomposition needs σ >= 1/4 for theoretical convergence"
        end
        return new{typeof(problem), length(M)}(problem, M, overlap, parallel, σ, ninner, subalg, workers)
    end
end

struct DualTVDDState{A,d,V,SAx,SC}
    algorithm::A

    "global variable"
    p::V
    "subdomain axes wrt global indices"
    subax::SAx
    "connectors to subworkers"
    cons::Array{SC,d}
end

function init(alg::DualTVDDAlgorithm{<:DualTVL1ROFOpProblem})
    g = alg.problem.g
    d = ndims(g)
    ax = axes(g)

    # subdomain axes
    subax = subaxes(size(g), alg.M, alg.overlap)
    # preallocated data for subproblems
    subg = [Array{eltype(g), d}(undef, length.(x)) for x in subax]
    # locally dependent tv parameter
    subλ = [alg.problem.λ[subax[i]...] .* theta.(Ref(ax), Ref(subax[i]), Ref(alg.overlap), CartesianIndices(subax[i]))
        for i in CartesianIndices(subax)]

    # TODO: generalize to SArray
    p1type(T::Type{<:Real}) = SVector{d,T}
    p1type(::Type{SVector{m,T}}) where {m, T} = SMatrix{m,d,T,m*d}

    # global dual variable
    p = zeros(p1type(eltype(g)), ax)
    # local dual variable
    subp = [zeros(p1type(eltype(g)), length.(sax)) for sax in subax]

    # create subproblem contexts
    cids = alg.parallel ? [eachindex(subax)] :
        chessboard_coloring(size(subax))
    cons = Array{Connector, d}(undef, size(subax))
    for (color, sidxs) in enumerate(cids)
        for (i, sidx) in enumerate(sidxs)
            sax = subax[sidx]

            subprob = DualTVL1ROFOpProblem(subg[sidx], op_restrict(alg.problem.B, ax, subax[sidx]), subλ[sidx])
            wf = subworker(alg, alg.subalg(subprob))
            wid = alg.workers[mod1(i, length(alg.workers))]
            cons[sidx] = outsource(wf, wid)
        end
    end

    return DualTVDDState(alg, p, subax, cons)
end

function intersectin(a, b)
    c = intersect(a, b)
    az = first(a) - one(eltype(a))
    bz = first(b) - one(eltype(b))
    return (c, c .- az, c .- bz)
end

function chessboard_coloring(sz)
    binli = LinearIndices(ntuple(_->2, length(sz)))
    coloring = [Int[] for _ in 1:2^length(sz)]

    li = LinearIndices(sz)
    for I in CartesianIndices(sz)
        push!(coloring[binli[CartesianIndex(mod1.(Tuple(I), 2))]], li[I])
    end

    return coloring
end

function subworker(alg, subalg)
    #fetch(st) .= reshape(reinterpret(SVector{d,eltype(g)}, initdata), size(g))

    ninner = alg.ninner
    return function(con)
        subst = init(subalg)
        while isopen(con)
            # fetch new data
            subg = take!(con)
            subalg.problem.g .= subg
            # run algorithm
            for _ in 1:ninner
                step!(subst)
            end
            # write result
            subp = fetch(subst)
            put!(con, subp)
        end
    end
end

function step!(ctx::DualTVDDState)
    alg = ctx.algorithm
    σ = ctx.algorithm.σ
    d = ndims(ctx.p)
    ax = axes(ctx.p)
    overlap = ctx.algorithm.overlap


    p_rem = copy(ctx.p)
    p_don = zeros(eltype(ctx.p), size(ctx.p))

    # subdomain loop (in coloring order)
    cids = alg.parallel ? [eachindex(ctx.subax)] :
        chessboard_coloring(size(ctx.subax))
    for (color, ids) in enumerate(cids)
        for pids in Iterators.partition(ids, length(alg.workers))
            for i in pids
                sax = ctx.subax[i]

                # update remaining old contribution
                view(p_rem, sax...) .-=
                    theta.(Ref(ax), Ref(sax), Ref(overlap), CartesianIndices(sax)) .* view(ctx.p, sax...)

                # data: g - Λ(p_don + p_rem)
                sg = copy(view(alg.problem.g, sax...))
                sp = extend(similar(ctx.p, length.(sax)), StaticKernels.ExtensionConstant(zero(eltype(ctx.p))))
                if alg.parallel
                    sp .= (1. .- theta.(Ref(ax), Ref(sax), Ref(overlap), CartesianIndices(sax))) .* view(ctx.p, sax...)
                else
                    sp .= view(p_don, sax...) .+ view(p_rem, sax...)
                end

                @inline kf(spw) = @inbounds sg[spw.position] + divergence(spw)
                kern = Kernel{ntuple(_->-1:1, d)}(kf)
                map!(kern, sg, sp)

                # start computation
                put!(ctx.cons[i], sg)
            end
            for i in pids
                sax = ctx.subax[i]

                sp = take!(ctx.cons[i])
                # reshape(reinterpret(SVector{d,eltype(alg.problem.g)}, ctx.subp[i]), size(ctx.subalg[i].problem.g))
                # weighted update for new contribution
                view(p_don, sax...) .+=
                    (1 .- σ) .* theta.(Ref(ax), Ref(sax), Ref(overlap), CartesianIndices(sax)) .* view(ctx.p, sax...) .+ σ .* sp
            end
        end
    end

    ctx.p .= p_don


    return ctx
end

fetch(ctx::DualTVDDState) = ctx.p

@generated function divergence_global(w::StaticKernels.Window{SVector{N,T},N}) where {N,T}
    i0 = ntuple(_->0, N)
    i1(k) = ntuple(i->Int(k==i), N)

    wi = (:(w[$(i0...)][$k] -
            (isnothing(w[$((.-i1(k))...)]) ? zero(T) : w[$((.-i1(k))...)][$k])) for k in 1:N)
    return quote
        Base.@_inline_meta
        return @inbounds +($(wi...))
    end
end

op_restrict(B::UniformScaling, ax, sax) = B
function op_restrict(B::Diagonal, ax, sax)
    Bdiag = diag(B)
    li = LinearIndices(ax)
    return Diagonal(vec(collect(Bdiag[li[i]] for i in CartesianIndices(sax))))
end
function op_restrict(B, ax, sax)
    li = LinearIndices(ax)
    si = vec(li[sax...])
    return B[si, si]
end


"""
    theta(ax, sax, overlap, i)

Return value of the partition function at index `i` given global axes `ax`,
subdomain axes `sax` and overlap count `overlap`.
This assumes that interior subdomains have size at least `2 .* overlap`.
"""
theta(ax, sax, overlap, i::CartesianIndex) = prod(theta.(ax, sax, overlap, Tuple(i)))
theta(ax, sax, overlap::Int, i::Int) =
    max(0., min(1.,
        first(ax) == first(sax) && i < first(ax) + overlap ? 1. : (i - first(sax)) / (overlap - 1),
        last(ax) == last(sax) && i > last(ax) - overlap ? 1. : (last(sax) - i) / (overlap - 1)))


"""
    subaxes(domain, pnum, overlap)

Determine axes for all subdomains, given per dimension number of domains
`pnum` and overlap `overlap`
"""
function subaxes(sizes, pnum, overlap)
    d = Val(length(sizes))
    ax = partition.(sizes, pnum, overlap)
    return [ntuple(i -> ax[i][I[i]], d) for I in CartesianIndices(pnum)]
end

function partition(n, k, o)
    part = UnitRange[]
    i = 1
    while k > 1
        s = (n + (k-1)*o) ÷ k
        push!(part, i:i+s-1)
        i = i + s - o
        n -= s - o
        k -= 1
    end
    push!(part, i:i+n-1)
end
